"""
Determine whether the given index is in bounds for a board
with `n` tiles.
"""
is_in_bounds(i::Integer, n_tiles::Integer) = 1 ≤ i ≤ n_tiles

"""
Check to see if the given value is either its max or min extrema
"""
is_type_extrema(v::T) where T <: Real = v == typemax(T) || v == typemin(T)

"""
col=1, row=1 will give us the top-left non-border square.
col=2, row=1 will give us the square just right of that, etc.
"""
function col_row_to_tile_index(board_width::Integer, col::Integer, row::Integer)::TileIndex
    return row*(board_width+2) + col + 1
end

"""
Get a tile's char representation.
"""
function square_to_char(v::TileValue)::Char
    retval = CH_FLOOR
    pbgw = v & (PLAYER+BOX+GOAL+WALL)
    if pbgw == PLAYER
        retval = CH_PLAYER
    elseif pbgw == PLAYER+GOAL
        retval = CH_PLAYER_ON_GOAL
    elseif pbgw == BOX || pbgw == PLAYER+BOX # can happen for deadlock-sets
        retval = CH_BOX
    elseif pbgw == BOX+GOAL || pbgw == PLAYER+BOX+GOAL # can happen for deadlock-sets
        retval = CH_BOX_ON_GOAL
    elseif pbgw == GOAL
        retval = CH_GOAL
    elseif pbgw == WALL
        retval = CH_WALL
    end
    return retval
end

function char_to_square(c::Char)::Union{TileValue,Nothing}
    if c == CH_BOX
        return BOX+FLOOR
    elseif c == CH_BOX_ON_GOAL
        return BOX+GOAL+FLOOR
    elseif c == CH_GOAL
        return GOAL+FLOOR
    elseif c == CH_FLOOR || c == CH_NON_BLANK_FLOOR
        return FLOOR
    elseif c == CH_PLAYER
        return PLAYER+FLOOR
    elseif c == CH_PLAYER_ON_GOAL
        return PLAYER+GOAL+FLOOR
    elseif c == CH_SQUARE_SET+FLOOR
        return FLAG_PLAYER_REACHABLE_SQUARE
    elseif c == CH_WALL
        return WALL
    else
        return nothing
    end
end

function dir_to_char(dir::Direction)::Char
    retval = '?'
    if dir == DIR_UP
        retval = '↑'
    elseif dir == DIR_LEFT
        retval = '←'
    elseif dir == DIR_DOWN
        retval = '↓'
    elseif dir == DIR_RIGHT
        retval = '→'
    end
    return retval
end

"""
Convert a dx and dy between two tiles to a direction
"""
function dxdy_to_direction(dx::Integer, dy::Integer)::Direction
    @assert(dx == 0 || dy == 0)
    if dx == 0
        return dy ≤ 0 ? DIR_UP : DIR_DOWN
    else
        return dx ≤ 0 ? DIR_LEFT : DIR_RIGHT
    end
end

"""
Goal tiles have their goal number (index in goals)
stored in their upper bits.
"""
function get_goal_number(v::TileValue)
    return v >> GOAL_BIT_SHIFT_COUNT
end

function count_tiles_with_flags(board::Board, flags::Int32)
    tot = 0
    for v in board
        tot += (v & flags) > 0
    end
    return tot
end

function is_a_legal_and_box_reachable_tile(v::TileValue)
    return v & (WALL + FLAG_ILLEGAL_BOX_TILE + FLAG_BOX_REACHABLE_TILE + FLOOR) == FLAG_BOX_REACHABLE_TILE + FLOOR
end

function print_char_matrix(chars)
    for i in 1:length(chars)
        for c in chars[i]
            print(c)
        end
        println("")
    end
end