"""
The sum of the distance to the nearest goal for all boxes,
which is a lower bound on the number of pushes (and thus also moves) to complete the board.
"""
function calculate_simple_lower_bound(s::State, dist_to_nearest_goal::Vector{Int32})::Int32
    value = zero(Int32)
    for □ in s.boxes
        if □ != 0 # box is on the board / hasn't been temporarily removed
            distance = dist_to_nearest_goal[□]
            if distance == typemax(Int32)
                # This box cannot reachy any goal
                return distance
            end
            value += distance
        end
    end
    return value
end


"""
Get the list of valid pushes for the given state.
NOTE: We assume reachability has already been computed.
NOTE: This method is not all that efficient - it allocates a moves vector.
"""
function get_pushes(game::Game, s::State, reach::ReachableTiles)
    pushes = Push[]
    for (box_number, □) in enumerate(s.boxes)
        if is_reachable_box(□, reach)
            # Get a move for each direction we can get at it from,
            # Provided that the opposing side is clear.
            for dir in DIRECTIONS
                □_dest = □ + game.step_fore[dir] # Destination tile
                □_player = □ - game.step_fore[dir] # Side of the player
                # If we can reach the player's pos and the destination is clear.
                if is_reachable(□_player, reach) && ((s.board[□_dest] & (FLOOR+BOX)) == FLOOR)
                    push!(pushes, Push(box_number, dir))
                end
            end
        end
    end
    return pushes
end

"""
Fill pushes with the list of valid pushes for the given state.
We return the number of valid pushes.
TODO: We assume reachability has already been computed.
"""
function get_pushes!(pushes::Vector{Push}, game::Game, s::State, reach::ReachableTiles)::Int
    n_pushes = 0
    for (box_number, □) in enumerate(s.boxes)
        if is_reachable_box(□, reach)
            # Get a move for each direction we can get at it from,
            # Provided that the opposing side is clear.
            for dir in DIRECTIONS
                □_dest = □ + game.step_fore[dir] # Destination tile
                □_player = □ - game.step_fore[dir] # Side of the player
                # If we can reach the player's pos and the destination is clear.
                if is_reachable(□_player, reach) && ((s.board[□_dest] & (FLOOR+BOX)) == FLOOR)
                    n_pushes += 1
                    pushes[n_pushes] = Push(box_number, dir)
                end
            end
        end
    end
    return n_pushes
end

"""
Convert a trajectory in pushes from the given start state to a trajectory in moves.
The given state should have the player in their correct starting position.
NOTE: We allocate memory.
"""
function convert_to_moves(game::Game, s0::State, pushes::Vector{Push})
    # Allocate memory
    s = deepcopy(s0)
    reach = ReachableTiles(length(s.board))
    clear!(reach, s.board)
    stack = Array{TileIndex}(undef, length(game.board))
    distances = zeros(Int32, length(s.board))
    queue = Array{Tuple{TileIndex, Int32}}(undef, length(s.board))

    moves = Direction[]
    for a in pushes
        ▩ = s.boxes[a.box_number] # box location
        □ = ▩ - game.step_fore[a.dir] # player push start location

        calc_reachable_tiles!(reach, game, s.board, s.player, stack)

        # Get distance to target
        player_orig = s.player
        s.player = □
        calc_distances!(distances, game, s, reach, queue)
        s.player = player_orig


        while s.player != □
            # Find a neighbor square that is closer.
            # We assume that it exists
            step_dir = typemax(Direction)
            ⬓ = zero(TileIndex)
            for dir in DIRECTIONS
                ◳ = s.player + game.step_fore[dir]
                if ⬓ == 0 || distances[◳] < distances[⬓]
                    ⬓ = ◳
                    step_dir = dir
                end
            end
            @assert step_dir != typemax(Direction)

            push!(moves, step_dir)
            move!(s, game, step_dir)
        end

        # Now apply the push
        push!(moves, a.dir)
        move!(s, game, a)
    end

    return moves
end