"""
This solver uses best-first search via the simple lower bound.
It also has a transposition table.
It differs from astar.jl in that it stores the full state in every node
of the search tree, and thus does not have to traverse the search tree
when switching between states. There is thus more memory allocation, but
less memory manipulation.
"""
struct AStarFullState
    d_max::Int
    timeout::Float64 # Maximum execution time in seconds
    n_positions::Int
end
AStarFullState(; d_max::Int=1, timeout::Float64=1.0, n_positions::Int=10) = AStarFullState(d_max, timeout, n_positions)

get_params(solver::AStarFullState) = Dict{String,Any}(
    "d_max"=>solver.d_max,
    "timeout"=>solver.timeout,
    "n_positions"=>solver.n_positions,
    )

mutable struct PositionFullState
    pred::PositionIndex # predecessor index in positions array
    push_depth::Int32
    player::TileIndex
    boxes::Vector{TileIndex}
    push::Push
end

function set_position!(
    s::State,
    game::Game,
    pos::PositionFullState,
    data::AStarData,
    )

    data.set_position_count_1 += 1

    remove_all_boxes_from_board!(s)
    s.boxes[:] = pos.boxes
    place_all_boxes_on_board!(s)

    move_player!(s, pos.player)
    s.zhash = calculate_zobrist_hash(s.boxes, game.hash_components)

    return s
end

function solve(solver::AStarFullState, game::Game, s::State)
    t_start = time()

    d_max, timeout, n_positions = solver.d_max, solver.timeout, solver.n_positions

    # Allocate memory and recalculate some things
    data = AStarData(false, 0, 0, 0, 0, 0, 0, 0, 0.0, 0.0, Array{Push}(undef, d_max), 0)

    reach = ReachableTiles(length(game.board))
    clear!(reach, game.board)
    stack = Array{TileIndex}(undef, length(game.board))

    queue_big = Array{Tuple{TileIndex, TileIndex, Int32}}(undef, length(game.board) * N_DIRS)
    distances_by_direction = Array{Int32}(undef, length(game.board), N_DIRS)
    dist_to_nearest_goal = zeros(Int32, length(game.board))
    calc_push_dist_to_nearest_goal_for_all_squares!(dist_to_nearest_goal, game, s, reach, stack, queue_big, distances_by_direction)

    # Calculate reachability and set player min pos
    calc_reachable_tiles!(reach, game, s.board, s.player, stack)
    move_player!(s, reach.min_reachable_tile)

    # Compute the lower bound
    simple_lower_bound = calculate_simple_lower_bound(s, dist_to_nearest_goal)
    # TODO: Return early if simple_lower_bound == 0

    pushes = Array{Push}(undef, 4*length(s.boxes))

    # Allocate a bunch of positions
    positions = Array{PositionFullState}(undef, n_positions)
    for i in 1:length(positions)
        positions[i] = PositionFullState(
            zero(PositionIndex),
            zero(Int32),
            zero(TileIndex),
            Array{TileIndex}(undef, length(s.boxes)),
            Push(0,0)
        )
    end

    # Index of the first free position
    # TODO: For now we don't re-use positions
    free_index = one(PositionIndex)

    # A priority queue used to order the states that we want to process,
    # in order by total cost (push_depth + lowerbound cost-to-go)
    to_process = PriorityQueue{PositionIndex, Int32}()

    # A dictionary that maps (player pos, zhash) to position index
    # TODO: This is a terrible way of doing this. Use the linked-list approach.
    closed_set = Dict{Tuple{TileIndex, UInt64}, PositionIndex}()

    # Store an infeasible state in the first index
    # We point to this position any time we get an infeasible state
    infeasible_state_index = free_index
    positions[infeasible_state_index].push_depth = zero(Int32)
    free_index += one(PositionIndex)

    # Store the root state in our set
    current_pos_index = free_index
    positions[free_index].pred = zero(Int32)
    positions[free_index].player = s.player
    positions[free_index].boxes[:] = s.boxes
    positions[free_index].push = Push(0,0)
    positions[free_index].push_depth = zero(Int32)
    free_index += one(PositionIndex)

    # Enqueue the root state
    enqueue!(to_process, current_pos_index, simple_lower_bound)
    closed_set[(s.player, s.zhash)] = current_pos_index

    data.setup_time = time() - t_start

    done = false
    while !done && !isempty(to_process) && (time() - t_start < timeout)
        pos_index = dequeue!(to_process)
        set_position!(s, game, positions[pos_index], data)
        push_depth = pos_index > 0 ? positions[pos_index].push_depth : zero(Int32)
        data.max_depth = max(data.max_depth, push_depth+1)
        calc_reachable_tiles!(reach, game, s.board, s.player, stack)

        # Iterate over pushes
        n_pushes = get_pushes!(pushes, game, s, reach)
        for push_index in 1:n_pushes
            push = pushes[push_index]

            data.n_pushes_evaled += 1
            move!(s, game, push)
            calc_reachable_tiles!(reach, game, s.board, s.player, stack)
            move_player!(s, reach.min_reachable_tile)

            # If we have not seen this state before so early, and it isn't an infeasible state.
            new_pos_index = get(closed_set, (s.player, s.zhash), zero(PositionIndex))
            if new_pos_index == 0 || push_depth+1 < positions[new_pos_index].push_depth

                lowerbound = calculate_simple_lower_bound(s, dist_to_nearest_goal)

                # Check to see if we are done
                if lowerbound == 0
                    # We are done!
                    done = true

                    # Back out the solution
                    data.solved = true
                    data.sol_depth = push_depth + 1
                    data.pushes[data.sol_depth] = push

                    pred = pos_index
                    while pred > 0
                        pos = positions[pred]
                        if pos.push_depth > 0
                            data.pushes[pos.push_depth] = pos.push
                        end
                        pred = pos.pred
                    end

                    break
                end

                # Enqueue and add to closed set if the lower bound indicates feasibility
                if lowerbound != typemax(Int32)

                    # Add the position
                    pos′ = positions[free_index]
                    pos′.pred = pos_index
                    pos′.player = s.player
                    pos′.boxes[:] = s.boxes
                    pos′.push = push
                    pos′.push_depth = push_depth + 1

                    to_process[free_index] = pos′.push_depth + lowerbound
                    closed_set[(s.player, s.zhash)] = free_index
                    free_index += one(PositionIndex)
                else
                    # Store infeasible state with a typemax index
                    data.n_infeasible_states += 1
                    closed_set[(s.player, s.zhash)] = infeasible_state_index
                end
            else
                data.n_pushes_skipped += 1
            end

            unmove!(s, game, push)
        end
    end

    data.n_positions_used = free_index - 1
    data.total_time = time() - t_start
    return data
end