"""
This solver uses A* via the simple lower bound.
It also has a transposition table that uses zobrist hashes.

This approach incorporates DFS expansions whenever a move maintains the same total cost.
That is, A* maintains a prority queue and always processes the node with the lowest estimated cost:
   cost = push_depth + simple_lower_bound
However, A* as to run set_position! to navigate to the new state, which costs some computation.
Whenever we do a push that makes progress toward our goals, we do not change the estimated cost.
   cost′ = (push_depth + 1) + (simple_lower_bound-1) = cost
We know that the previous cost was the lowest cost in our priotiy queue, so the new state must also
be the lowest cost in our priority queue. We can thus continue on with it and avoid computation.
See "Combining depth-first search with the open-list in A*" at
http://sokobano.de/wiki/index.php?title=Sokoban_solver_%22scribbles%22_by_Brian_Damgaard_about_the_YASS_solver#Goal_room_packing
"""
@kwdef struct AStarWithDFSExpansion
    d_max::Int = 1
    timeout::Float64 = 1.0 # Maximum execution time in seconds
    n_positions::Int = 10
end

get_params(solver::AStarWithDFSExpansion) = Dict{String,Any}(
    "d_max"=>solver.d_max,
    "timeout"=>solver.timeout,
    "n_positions"=>solver.n_positions,
    )

mutable struct AStarWithDFSExpansionMemory
    free_index::PositionIndex
end

mutable struct AStarWithDFSExpansionData
    solved::Bool
    max_depth::Int # max depth reached during search
    n_pushes_evaled::Int # number of pushes evaluated during search
    n_pushes_skipped::Int # number of pushes skipped due to the transposition table
    n_dfs_expansions::Int
    n_infeasible_states::Int
    set_position_count_1::Int
    set_position_count_2::Int
    n_positions_used::Int
    setup_time::Float64 # time it takes to set up the problem
    total_time::Float64
    pushes::Vector{Push} # of length d_max
    sol_depth::Int # number of pushes in the solution
end

function print_results(data::AStarWithDFSExpansionData, verbosity::Int=1)
    println("\t\tsolved: ", data.solved)
    println("\t\ttotal time: ", data.total_time)
    if verbosity > 0
        println("\t\tsetup_time: ", data.setup_time, " ($(round(data.setup_time/data.total_time*100,digits=2))%)")
        println("\t\tsolving time: ", data.total_time - data.setup_time)
        println("\t\tmax_depth: ", data.max_depth)
        println("\t\tsol_depth: ", data.sol_depth)
        println("\t\tn_pushes_evaled: ", data.n_pushes_evaled)
        println("\t\tn_pushes_skipped: ", data.n_pushes_skipped)
        println("\t\tn_dfs_expansions: ", data.n_dfs_expansions)
        println("\t\tn_infeasible_states: ", data.n_infeasible_states)
        println("\t\tset_position_count_1: ", data.set_position_count_1)
        println("\t\tset_position_count_2: ", data.set_position_count_2)
        println("\t\tn_positions_used: ", data.n_positions_used)
    end
end

function expand_node!(
    mem::AStarWithDFSExpansionMemory,
    data::AStarWithDFSExpansionData,
    game::Game,
    s::State,
    stack::Vector{TileIndex},
    reaches::Vector{ReachableTiles},
    dist_to_nearest_goal::Vector{Int32},
    positions::Vector{Position},
    to_process::PriorityQueue{PositionIndex, Int32},
    closed_set::Dict{Tuple{TileIndex, UInt64}, PositionIndex},
    pos_index::PositionIndex,
    push_depth::Int32,
    lowerbound::Int32, # Current lowerbound value
    reach_index::Int,
    )

    # NOTE: We expect that reachability was already computed for the given reach index

    prev_cost = push_depth + lowerbound
    data.max_depth = max(data.max_depth, push_depth + 1)

    # Iterate over pushes
    for box_number in 1:length(s.boxes)
        □ = s.boxes[box_number]
        if is_reachable_box(□, reaches[reach_index])

            # Get a move for each direction we can get at it from,
            # Provided that the opposing side is clear.
            for dir in DIRECTIONS
                □_dest = □ + game.step_fore[dir] # Destination tile
                □_player = □ - game.step_fore[dir] # Side of the player
                # If we can reach the player's pos and the destination is clear.
                if is_reachable(□_player, reaches[reach_index]) && ((s.board[□_dest] & (FLOOR+BOX)) == FLOOR)
                    push = Push(box_number, dir)
                    data.n_pushes_evaled += 1


                    # Move, with an incremental update of the lowerbound
                    # NOTE: We are guaranteed to only push a box with a finite dist-to-nearest-goal
                    lowerbound_new = lowerbound - dist_to_nearest_goal[s.boxes[push.box_number]]
                    move!(s, game, push)
                    lowerbound_component = dist_to_nearest_goal[s.boxes[push.box_number]]
                    if lowerbound_component == typemax(Int32)
                        # Infeasible!
                        # Do not proceed
                        data.n_infeasible_states += 1
                    else
                        lowerbound_new += lowerbound_component
                        calc_reachable_tiles!(reaches[reach_index+1], game, s.board, s.player, stack)
                        move_player!(s, reaches[reach_index+1].min_reachable_tile)

                        # If we have not seen this (feasible) state before so early
                        new_pos_index = get(closed_set, (s.player, s.zhash), zero(PositionIndex))
                        if new_pos_index == 0 || push_depth+one(Int32) < positions[new_pos_index].push_depth

                            # Check to see if we are done
                            if lowerbound_new == 0
                                # We are done!
                                # Back out the solution
                                data.solved = true
                                data.sol_depth = push_depth + one(Int32)
                                data.pushes[data.sol_depth] = push

                                pred = pos_index
                                while pred > 0
                                    pos = positions[pred]
                                    data.pushes[pos.push_depth] = pos.push
                                    pred = pos.pred
                                end

                                return
                            end

                            # Add the position
                            pos_index′ = mem.free_index
                            mem.free_index += one(PositionIndex)

                            pos′ = positions[pos_index′]
                            pos′.pred = pos_index
                            pos′.push = push
                            pos′.push_depth = push_depth + one(Int32)
                            pos′.lowerbound = lowerbound_new

                            # Add to the closed set
                            closed_set[( s.player, s.zhash)] = pos_index′

                            new_cost = pos′.push_depth + pos′.lowerbound
                            if new_cost == prev_cost
                                # If the new position has the same score as its parent, we think it is as good as current node,
                                # which means it is best priority, so might as well expand now and save computation.
                                expand_node!(mem, data, game, s, stack, reaches, dist_to_nearest_goal, positions, to_process, closed_set, pos_index′, pos′.push_depth, pos′.lowerbound, reach_index+1)
                                if data.solved
                                    return
                                end
                            else
                                # Add the new position to the priority queue, using the proper score
                                to_process[pos_index′] = new_cost
                            end
                        else
                            data.n_pushes_skipped += 1
                        end
                    end

                    unmove!(s, game, push)
                end
            end
        end
    end
end

function solve(solver::AStarWithDFSExpansion, game::Game, s::State)
    t_start = time()

    d_max, timeout, n_positions = solver.d_max, solver.timeout, solver.n_positions

    # Allocate memory and recalculate some things
    data = AStarWithDFSExpansionData(false, 0, 0, 0, 0, 0, 0, 0, 0, 0.0, 0.0, Array{Push}(undef, d_max), 0)

    stack = Array{TileIndex}(undef, length(game.board))
    reaches = Array{ReachableTiles}(undef, d_max)
    for d in 1:d_max
        reaches[d] = ReachableTiles(length(game.board))
        clear!(reaches[d], game.board)
    end

    # Calculate reachability and set player min pos
    calc_reachable_tiles!(reaches[1], game, s.board, s.player, stack)
    move_player!(s, reaches[1].min_reachable_tile)

    # Calculate distance to nearest goal for all squares
    queue_big = Array{Tuple{TileIndex, TileIndex, Int32}}(undef, length(s.board) * N_DIRS)
    distances_by_direction = Array{Int32}(undef, length(s.board), N_DIRS)
    dist_to_nearest_goal = zeros(Int32, length(s.board))
    calc_push_dist_to_nearest_goal_for_all_squares!(dist_to_nearest_goal, game, s, reaches[1], stack, queue_big, distances_by_direction)

    # Compute the lower bound
    simple_lower_bound = calculate_simple_lower_bound(s, dist_to_nearest_goal)
    # TODO: Return early if simple_lower_bound == 0

    pushes = Array{Push}(undef, 4*length(s.boxes))

    # Allocate a bunch of positions
    positions = Array{Position}(undef, n_positions)
    for i in 1:length(positions)
        positions[i] = Position(
            zero(TileIndex),
            Push(0,0),
            zero(Int32),
            zero(Int32),
            false
        )
    end

    mem = AStarWithDFSExpansionMemory(
        one(PositionIndex) # Index of the first free position, TODO: For now we don't re-use positions
    );

    # A priority queue used to order the states that we want to process,
    # in order by total cost (push_depth + lowerbound cost-to-go)
    to_process = PriorityQueue{PositionIndex, Int32}()

    # A dictionary that maps (player pos, zhash) to position index
    closed_set = Dict{Tuple{TileIndex, UInt64}, PositionIndex}()

    # Enqueue the root state
    current_pos_index = zero(PositionIndex)
    enqueue!(to_process, current_pos_index, simple_lower_bound)
    closed_set[(s.player, s.zhash)] = current_pos_index

    data.setup_time = time() - t_start

    while !data.solved && !isempty(to_process) && (time() - t_start < timeout)
        pos_index = dequeue!(to_process)
        set_position!(s, game, positions, pos_index, current_pos_index, data)
        current_pos_index = pos_index

        push_depth = zero(Int32)
        lowerbound = simple_lower_bound
        if pos_index > 0
            pos = positions[pos_index]
            push_depth = pos.push_depth
            lowerbound = pos.lowerbound
        end

        calc_reachable_tiles!(reaches[1], game, s.board, s.player, stack)
        expand_node!(mem, data, game, s, stack, reaches, dist_to_nearest_goal, positions, to_process, closed_set, pos_index, push_depth, lowerbound, 1)
    end

    data.n_positions_used = mem.free_index - 1
    data.total_time = time() - t_start
    return data
end