# DFS with a transposition table using Zobrist hash keys
# If a state has been seen before, skip it
struct ZobristDFS
    d_max::Int
    stack::Vector{TileIndex} # Memory used for reachability calcs
    transposition_table::Dict{Tuple{TileIndex, UInt64},Int} # (player normalized pos, box zhash) → d_min
end

mutable struct ZobristDFSData
    solved::Bool
    max_depth::Int # max depth reached during search
    n_pushes_evaled::Int # number of pushes evaluated during search
    n_pushes_skipped::Int # number of pushes skipped due to the transposition table
    solve_time::Float64 # solve time in seconds
    pushes::Vector{Push} # of length d_max
    sol_depth::Int # number of pushes in the solution
end

function clear!(data::ZobristDFSData)
    data.solved = false
    data.max_depth = 0
    data.n_pushes_evaled = 0
    data.n_pushes_skipped = 0
    data.solve_time = 0.0
    data.sol_depth = 0
    return data
end

function solve_(dfs::ZobristDFS, data::ZobristDFSData, game::Game, s::State, reach::ReachableTiles, d::Int)

    data.max_depth = max(data.max_depth, d)

    if is_solved(game, s)
        # Solved!
        data.solved = true
        data.sol_depth = d
        return true
    end

    if d ≥ dfs.d_max
        # We exceeded our max depth
        return false
    end

    # Calc reachability for this state
    calc_reachable_tiles!(reach, game, s.board, s.player, dfs.stack)

    # Set the position to the normalized location
    s.player = reach.min_reachable_tile

    # Exit early if we have seen this state before at this shallow of a depth
    ztup = (s.player, s.zhash)
    d_seen = get(dfs.transposition_table, ztup, typemax(Int))
    if d_seen ≤ d
        # Skip!
        data.n_pushes_skipped += 1
        return false
    end

    # Store this state hash in our transposition table
    dfs.transposition_table[ztup] = d

    for push in get_pushes(game, s, reach)
        data.n_pushes_evaled += 1
        move!(s, game, push)
        if solve_(dfs, data, game, s, reach, d+1)
            # Store the solution
            data.pushes[d+1] = push
            return true
        end
        unmove!(s, game, push)
    end

    # Failed to find a solution
    return false
end

function solve(dfs::ZobristDFS, game::Game, s::State, reach::ReachableTiles)
    t_start = time()
    empty!(dfs.transposition_table) # Ensure we always start with an empty transposition table
    data = ZobristDFSData(false, 0, 0, 0, 0.0, Array{Push}(undef, dfs.d_max), 0)
    clear!(reach, game.board)
    solve_(dfs, data, game, s, reach, 0)
    data.solve_time = time() - t_start
    return data
end