using Luxor
using Colors

const COLOR_WALL = (0.2, 0.2, 0.2)
const COLOR_FLOOR = (0.4, 0.4, 0.4)
const COLOR_TILE = (0.5, 0.5, 0.5)
const COLOR_GOAL = (0x1B/0xFF, 0xA1/0xFF, 0xEA/0xFF)
const COLOR_PLAYER = (0xF5/0xFF, 0x61/0xFF, 0x5C/0xFF)
const COLOR_BOX = (0x70/0xFF, 0xA1/0xFF, 0x7C/0xFF)
const COLOR_MOVE = (0xF5/0xFF, 0xCE/0xFF, 0x42/0xFF)

const PIX_PER_SIDE = 25

function draw_walls_and_floors(game::Game, board::Board)
    d = Drawing(
        (game.board_width+2)*PIX_PER_SIDE,
        (game.board_height+2)*PIX_PER_SIDE,
        :png)

    # Render all walls and floors
    for row in 0:game.board_height+1
        for col in 0:game.board_width+1
            x = col*PIX_PER_SIDE
            y = row*PIX_PER_SIDE
            v = board[col_row_to_tile_index(game, col, row)]
            if (v & WALL) > 0
                sethue(COLOR_WALL...)
                rect(x, y, PIX_PER_SIDE, PIX_PER_SIDE, :fill)
            end
            if (v & FLOOR) > 0
                sethue(COLOR_FLOOR...)
                rect(x, y, PIX_PER_SIDE, PIX_PER_SIDE, :fill)
                sethue(COLOR_TILE...)
                rect(x + 2, y + 2, PIX_PER_SIDE - 4, PIX_PER_SIDE - 4, :fill)
            end
        end
    end

    return d
end

function draw_board!(d::Drawing, game::Game, board::Board)
    # Render all goals
    sethue(COLOR_GOAL...)
    for □ in game.goals
        col, row = tile_index_to_col_row(game, □)
        x = (col+0.5)*PIX_PER_SIDE
        y = (row+0.5)*PIX_PER_SIDE
        circle(x, y, PIX_PER_SIDE/4, :fill)
    end

    # Render boxes and player
    for row in 0:game.board_height+1
        for col in 0:game.board_width+1
            x = (col+0.5)*PIX_PER_SIDE
            y = (row+0.5)*PIX_PER_SIDE
            v = board[col_row_to_tile_index(game, col, row)]
            if (v & BOX) > 0
                # color according to whether it is on goal or not
                if (v & GOAL) > 0
                    sethue(COLOR_GOAL...)
                else
                    sethue(COLOR_BOX...)
                end
                box(Point(x, y), PIX_PER_SIDE-2, PIX_PER_SIDE-2, 5, :fill)
            end
            if (v & PLAYER) > 0
                sethue(COLOR_PLAYER...)
                star(Point(x,y), PIX_PER_SIDE/3, 5, 0.5, 0.0, :fill)
            end
        end
    end

    return d
end

function draw_board(game::Game, board::Board)
    d = draw_walls_and_floors(game, board)
    draw_board!(d, game, board)
    finish()
    d
end

function draw_board_flag(game::Game, flags::Integer)
    d = draw_walls_and_floors(game, game.board)

    for row in 0:game.board_height+1
        for col in 0:game.board_width+1
            x = col*PIX_PER_SIDE
            y = row*PIX_PER_SIDE
            □ = col_row_to_tile_index(game, col, row)
            vf = game.board[□] & flags
            if vf > 0
                hue = hash(vf) % 360
                rgb = convert(RGB, HSL(hue, 0.75, 0.75))
                setcolor(rgb.r, rgb.g, rgb.b, 0.5)
                rect(x, y, PIX_PER_SIDE, PIX_PER_SIDE, :fill)
            end
        end
    end

    finish()
    d
end

function draw_reachability(game::Game, reach::ReachableTiles)
    d = draw_walls_and_floors(game, game.board)

    # Render reachability
    for row in 0:game.board_height+1
        for col in 0:game.board_width+1
            x = col*PIX_PER_SIDE
            y = row*PIX_PER_SIDE
            v = reach.tiles[col_row_to_tile_index(game, col, row)]
            if reach.calc_counter ≤ v ≤ reach.calc_counter + 1
                if v == reach.calc_counter
                    # reachable square
                    setcolor(COLOR_PLAYER..., 0.5)
                else
                    # neighboring box
                    setcolor(COLOR_BOX..., 0.5)
                end
                rect(x, y, PIX_PER_SIDE, PIX_PER_SIDE, :fill)
            end
        end
    end

    # Render min reachable tile
    let
        col, row = tile_index_to_col_row(game, reach.min_reachable_tile)
        x = col*PIX_PER_SIDE
        y = row*PIX_PER_SIDE
        sethue(COLOR_PLAYER)
        rect(x, y, PIX_PER_SIDE, PIX_PER_SIDE, :fill)
    end

    finish()
    d
end

function draw_distances!(d::Drawing, game::Game, distances::Vector{Int32})
    # Render distances
    color_lo = COLOR_PLAYER
    color_hi = COLOR_GOAL
    max_non_inf_distance = zero(Int32)
    for dist in distances
        if dist < typemax(Int32)
            max_non_inf_distance = max(max_non_inf_distance, dist)
        end
    end

    for row in 0:game.board_height+1
        for col in 0:game.board_width+1
            x = col*PIX_PER_SIDE
            y = row*PIX_PER_SIDE
            dist = distances[col_row_to_tile_index(game, col, row)]
            if dist < typemax(Int32)
                a = dist/max_non_inf_distance
                r = color_lo[1]*(1.0 - a) + color_hi[1]*a
                g = color_lo[2]*(1.0 - a) + color_hi[2]*a
                b = color_lo[3]*(1.0 - a) + color_hi[3]*a
                sethue(r, g, b)
                rect(x, y, PIX_PER_SIDE, PIX_PER_SIDE, :fill)
            end
        end
    end

    return d
end

function draw_distances(game::Game, distances::Vector{Int32})
    d = draw_walls_and_floors(game, game.board)
    draw_distances!(d, game, distances)
    finish()
    d
end

function draw_distances_and_board(game::Game, s::State, distances::Vector{Int32})
    d = draw_walls_and_floors(game, s.board)
    draw_distances!(d, game, distances)
    draw_board!(d, game, s.board)
    finish()
    d
end

function draw_distances_by_direction!(d::Drawing, game::Game, distances_by_direction::Matrix{Int32})
    # Render distances by direction
    color_lo = COLOR_PLAYER
    color_hi = COLOR_GOAL
    max_non_inf_distance = zero(Int32)
    for dist in distances_by_direction
        if dist < typemax(Int32)
            max_non_inf_distance = max(max_non_inf_distance, dist)
        end
    end

    # Render a color triangle for each tile.
    # Triangles are only rendered if distances are finite.
    for row in 0:game.board_height+1
        for col in 0:game.board_width+1
            □ = col_row_to_tile_index(game, col, row)
            x = col*PIX_PER_SIDE
            y = row*PIX_PER_SIDE

            tl = Point(x,y)
            tr = Point(x+PIX_PER_SIDE,y)
            bl = Point(x,y+PIX_PER_SIDE)
            br = Point(x+PIX_PER_SIDE,y+PIX_PER_SIDE)
            center = Point(x+PIX_PER_SIDE/2.0, y+PIX_PER_SIDE/2.0)

            for (dir, triangle) in [
                    (DIR_UP,    [center, tr, tl, center]),
                    (DIR_RIGHT, [center, br, tr, center]),
                    (DIR_DOWN,  [center, bl, br, center]),
                    (DIR_LEFT,  [center, tl, bl, center]),
                ]
                dist = distances_by_direction[□, dir]
                if !is_type_extrema(dist)
                    a = dist/max_non_inf_distance
                    r = color_lo[1]*(1.0 - a) + color_hi[1]*a
                    g = color_lo[2]*(1.0 - a) + color_hi[2]*a
                    b = color_lo[3]*(1.0 - a) + color_hi[3]*a
                    sethue(r, g, b)
                    poly(triangle, :fill)
                end
            end
        end
    end

    return d
end

function draw_distances_by_direction(game::Game, distances_by_direction::Matrix{Int32})
    d = draw_walls_and_floors(game, game.board)
    draw_distances_by_direction!(d, game, distances_by_direction)
    finish()
    d
end

"""
Draw a set of available moves for a single state.
"""
function draw_pushes(game::Game, s::State, pushes::Vector{Push})
    d = draw_walls_and_floors(game, s.board)
    draw_board!(d, game, s.board)

    # Render pushes
    sethue(COLOR_MOVE)
    for push in pushes
        □ = s.boxes[push.box_number]
        col, row = tile_index_to_col_row(game, □)
        x1 = (col+0.5)*PIX_PER_SIDE
        y1 = (row+0.5)*PIX_PER_SIDE

        ▩ = □ + game.step_fore[push.dir]
        col, row = tile_index_to_col_row(game, ▩)
        x2 = (col+0.5)*PIX_PER_SIDE
        y2 = (row+0.5)*PIX_PER_SIDE
        arrow(Point(x1, y1), Point(x2, y2))
    end

    finish()
    d
end


"""
Render a game sequence starting from initial state `s0` and applying the moves
in `moves` in order.
The moves are all assumed to be valid.
"""
function draw_trajectory(game::Game, s0::State, pushes::Vector{Push})
    s = deepcopy(s0)
    sol_depth = length(pushes)
    frames = Array{Drawing}(undef, 1+sol_depth)
    frames[1] = draw_board(game, s.board)
    for d in 1:sol_depth
        move!(s, game, pushes[d])
        frames[d+1] = draw_board(game, s.board)
    end
    return frames
end

function draw_trajectory(game::Game, s0::State, moves::Vector{Direction})
    s = deepcopy(s0)
    sol_depth = length(moves)
    frames = Array{Drawing}(undef, 1+sol_depth)
    frames[1] = draw_board(game, s.board)
    for d in 1:sol_depth
        move!(s, game, moves[d])
        frames[d+1] = draw_board(game, s.board)
    end
    return frames
end


function draw_board_isometric(game::Game, board::Board)

    dx = PIX_PER_SIDE*cosd(30)
    dy = PIX_PER_SIDE*sind(30)

    # Vector as rows increase
    d_row = Point(dx,  dy)

    # Vector as columns increase
    d_col = Point(dx, -dy)

    center = Point(2*PIX_PER_SIDE, 2*PIX_PER_SIDE)
    pts_tile = [
        Point(-dx,0),
        Point(0,dy),
        Point(dx,0),
        Point(0,-dy),
        Point(-dx,0)
    ]
    pts_tile_bot_left = [
        Point(-dx,    0),
        Point(-dx,1.75dy),
        Point(  0,2.75dy),
        Point(  0,    dy),
        Point(-dx,    0),
    ]
    pts_tile_bot_right = [
        Point(  0,    dy),
        Point(  0,2.75dy),
        Point( dx,1.75dy),
        Point( dx,     0),
        Point(  0,    dy),
    ]
    pts_goal = pts_tile .* 0.5
    pts_box_top = [
        Point(-dx,-2dy),
        Point(  0, -dy),
        Point( dx,-2dy),
        Point(  0,-3dy),
        Point(-dx,-2dy)
    ]
    box_inset = 0.75
    pts_box_top_inset = [
        Point(-dx*box_inset,-2dy),
        Point(  0, -(2-box_inset)*dy),
        Point( dx*box_inset,-2dy),
        Point(  0,-(2+box_inset)*dy),
        Point(-dx*box_inset,-2dy)
    ]
    pts_box_left = [
        Point(-dx,   0),
        Point(  0,  dy),
        Point(  0, -dy),
        Point(-dx,-2dy),
        Point(-dx,   0)
    ]
    pts_box_right = [
        Point(  0,  dy),
        Point( dx,   0),
        Point( dx,-2dy),
        Point(  0, -dy),
        Point(  0,  dy)
    ]

    # Create the drawing
    buffer = 10 # pixels
    x_min = -dx - buffer
    x_max = dx*(game.board_width+game.board_height+3) + buffer
    y_min = -dy*(game.board_width+2) - buffer
    y_max = dy*(game.board_height+4) + buffer
    offset = Point(-x_min, -y_min)

    d = Drawing(ceil(Int, x_max - x_min), ceil(Int, y_max - y_min), :png)
    Luxor.background(COLOR_WALL...)

    # Get the tiles in order
    ordered_tiles = Tuple{Int,Int,TileIndex}[]
    col_max = game.board_width+1
    row_max = game.board_height+1
    for col_start in col_max : - 1 : 0
        for row in 0 : min(row_max, col_max-col_start)
            col = col_start + row
            □ = col_row_to_tile_index(game, col, row)
            push!(ordered_tiles, (row, col, □))
        end
    end
    for row_start in 1:row_max
        for col in 0 : min(col_max, row_max-row_start)
            row = row_start + col
            □ = col_row_to_tile_index(game, col, row)
            push!(ordered_tiles, (row, col, □))
        end
    end

    # Render all floor tiles, back to front
    COLOR_TILE_LEFT = COLOR_TILE.*0.75
    COLOR_TILE_RIGHT = COLOR_TILE_LEFT.*0.75
    for (row, col, □) in ordered_tiles
        v = board[□]
        if (v & FLOOR) > 0
            gsave()
            translate(row*d_row + col*d_col + offset)
            sethue(COLOR_TILE_LEFT...)
            poly(pts_tile_bot_left, :fill)
            sethue(COLOR_TILE_RIGHT...)
            poly(pts_tile_bot_right, :fill)
            sethue(COLOR_TILE...)
            poly(pts_tile, :fill)
            grestore()
        end
    end

    # Render all goals
    sethue(COLOR_GOAL...)
    for □ in game.goals
        col, row = tile_index_to_col_row(game, □)
        gsave()
        translate(row*d_row + col*d_col + offset)
        poly(pts_goal, :stroke)
        grestore()
    end

    # Render boxes and players, back to front
    COLOR_BOX = (0xBE/0xFF, 0xFA/0xFF, 0xFA/0xFF)
    COLOR_BOX_ON_GOAL = (0x19/0xFF, 0xAE/0xFF, 0xFF/0xFF)
    for (row, col, □) in ordered_tiles
        v = board[□]
        if (v & BOX) > 0
            color_box = (v & GOAL) > 0 ? COLOR_BOX_ON_GOAL : COLOR_BOX
            color_box_left = color_box.*0.75
            color_box_right = color_box_left.*0.75

            gsave()
            translate(row*d_row + col*d_col + offset)
            sethue(color_box...)
            poly(pts_box_top, :fill)
            sethue(color_box_left...)
            poly(pts_box_left, :fill)
            sethue(color_box_right...)
            poly(pts_box_right, :fill)
            grestore()
        end

        if (v & PLAYER) > 0
            gsave()
            translate(row*d_row + col*d_col + offset)
            sethue(COLOR_PLAYER...)
            circle(0, -dy, PIX_PER_SIDE/2, :fill)
            grestore()
        end
    end

    finish()

    d
end

function draw_trajectory_isometric(game::Game, s0::State, moves::Vector{Direction})
    s = deepcopy(s0)
    sol_depth = length(moves)
    frames = Array{Drawing}(undef, 1+sol_depth)
    frames[1] = draw_board_isometric(game, s.board)
    for d in 1:sol_depth
        move!(s, game, moves[d])
        frames[d+1] = draw_board_isometric(game, s.board)
    end
    return frames
end